/*
 * dskel.c
 *
 *  Created on: 14.6.2011
 *      Author: emlin
 */

#include "dskel-ioctl.h"

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/kernel.h>

#include <asm/uaccess.h>

#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/jiffies.h>
#include <linux/timer.h>

// -----------------------------------------------------------------------------
#define DEV_NAME					"dskel"
#define DEV_COUNT						1

#define PAR_PORT_BASE				0x378
#define PAR_PORT_SIZE				3
// ofset od zacatku bloku registru paralelniho portu
#define PAR_PORT_REG_DATA			(0x00)
#define PAR_PORT_REG_STAT			(0x01)
#define PAR_PORT_REG_CTRL			(0x02)

// -----------------------------------------------------------------------------
enum
{
	ModuleExit = 0,

	Failed_ChrdevRegion,
	Failed_CdevAlloc,
	Failed_CdevAdd,
	Failed_SysClass,
	Failed_SysDev,
	Failed_ReqRegion,
	Failed_IoMap,
};

// -----------------------------------------------------------------------------
// parametry ovladace
struct dskel_t
{
	// major-minor par
	dev_t				dev_num;
	// ukazatel na strukturu cdev reprezentujici znakove zarizeni
	struct cdev *		cdev_ptr;
	// ukazatel na tridu zarizeni v adresari /sys/class
	struct class *		sys_class_ptr;
	struct device * 	sys_dev_ptr;
	// hlida pocet volani operace open()
	atomic_t			open_cnt;
	// zacatek bloku registru paralelniho portu namapovaneho jako I/O pamet
	void __iomem *		port_base;
	// casovac
	struct timer_list	timer;
	// perioda casovace (milisec)
	atomic_t 			timer_period;
	// promenna pro deaktivaci casovace
	// (1 = deaktivuj casovac, 0 = casovac byl aktivovan
	atomic_t 			timer_stop;
};

static struct dskel_t dskel =
{
	.dev_num = MKDEV(0, 0),
	.cdev_ptr = NULL,
	.sys_class_ptr = NULL,
	.sys_dev_ptr = NULL,
	.open_cnt = ATOMIC_INIT(0),
	.port_base = NULL,
	.timer_period = ATOMIC_INIT(1000),
	.timer_stop = ATOMIC_INIT(1),
};

// -----------------------------------------------------------------------------
static void timer_on(unsigned int period)
{
	// vynulujeme si deaktivacni promenou
	atomic_set(&dskel.timer_stop, 0);

	// zapamatujeme si periodu (milisecs)
	atomic_set(&dskel.timer_period, period);

	// nastavime si znacku, kdy ma casovac vyprset
	dskel.timer.expires = jiffies + msecs_to_jiffies(period);

	// spustime casovac
	add_timer(&dskel.timer);
}

// -----------------------------------------------------------------------------
static void timer_off(void)
{
	atomic_set(&dskel.timer_stop, 1);

	del_timer(&dskel.timer);
}

// -----------------------------------------------------------------------------
void timer_userfunction(unsigned long data)
{
	/* nejaky uzitecny kod */
	pr_alert("dskel: timer_userfunction()\n");

	if ( !atomic_read(&dskel.timer_stop) )
	{
		mod_timer(&dskel.timer,
				  jiffies + msecs_to_jiffies(atomic_read(&dskel.timer_period)));
	}
}

// -----------------------------------------------------------------------------
static int dskel_open(struct inode * inode_ptr, struct file * file_ptr)
{
	int result = 0, tmp_open_cnt = 0;

	// soubor zarizeni muze byt otevren jen jednim programem soucasne
	// atomicky inkrementne promenou o 1 a vrati PUVODNI hodnotu
	tmp_open_cnt = atomic_inc_and_test(&dskel.open_cnt);
	if ( tmp_open_cnt > 1 )
	{
		atomic_dec(&dskel.open_cnt);
		result = -EBUSY;
		goto End;
	}

	// modul nepodporuje systemove volani lseek()
	result = nonseekable_open(inode_ptr, file_ptr);
	if ( result < 0 )
	{
		goto End;
	}

End:
	pr_alert("dskel_open(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int dskel_release(struct inode * inode_ptr, struct file * file_ptr)
{
	// povolime nove otevreni
	atomic_dec(&dskel.open_cnt);

	pr_alert("dskel_release(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
static ssize_t dskel_read(struct file * file_ptr,
				   		  char __user * data_ptr, size_t data_size,
				   		  loff_t * pos_ptr)
{
	pr_alert("dskel_read(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
static ssize_t dskel_write(struct file * file_ptr,
				    const char __user * data_ptr, size_t data_size,
				    loff_t * pos_ptr)
{
	pr_alert("dskel_write(): %d\n", data_size);

	return data_size;
}

// -----------------------------------------------------------------------------
static int dskel_ioctl(struct inode * inode_ptr, struct file * file_ptr,
				       unsigned int cmd, unsigned long arg)
{
	int result = 0, tmp_var = 0;

	// kontrola kodu prikazu
	if ( (_IOC_TYPE(cmd) != DSKEL_MAGIC_NO) || (_IOC_NR(cmd) > DSKEL_MAX_CMD_NR) )
	{
		result = -ENOTTY;
		goto End;
	}

	// kontrola argumentu (platnost ukazatele a smeru pristupu)
	// Pozn.: smery v access_ok() jsou presne opacne nez u ioctl
	if ( _IOC_DIR(cmd) & _IOC_READ )
	{
		result = access_ok(VERIFY_WRITE, (void __user *) arg, _IOC_SIZE(cmd));
	}
	else if ( _IOC_DIR(cmd) & _IOC_WRITE )
	{
		result = access_ok(VERIFY_READ, (void __user *) arg, _IOC_SIZE(cmd));
	}
	else
	{ ; /*empty*/ }

	if ( result < 0 )
	{
		goto End;
	}

	// dekodujeme prikaz a pak jej provedeme
	switch( cmd )
	{
		case DSKEL_GET_PARAM:
			result = __put_user(tmp_var, (int *) arg);
			break;

		case DSKEL_SET_PARAM:
			result = __get_user(tmp_var, (int *) arg);
			break;

		case DSKEL_TIMER_ON:
		{
			result = __get_user(tmp_var, (int *) arg);
			if ( result < 0 )
			{
				goto End;
			}

			timer_on(tmp_var);
		}
		break;

		case DSKEL_TIMER_OFF:
			timer_off();
			break;

		default:
			result = -ENOTTY;
			break;
	}

End:
	pr_alert("dskel_ioctl(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
// virtualni tabulka - tabulka operaci podporovanych ovladacem
static struct file_operations dskel_file_ops =
{
	.owner = THIS_MODULE,
	.open = &dskel_open,
	.release = &dskel_release,
	.read = &dskel_read,
	.write = &dskel_write,
	.ioctl = &dskel_ioctl,
	// modul nepodporuje metodu llseek
	.llseek = &no_llseek,
};

// -----------------------------------------------------------------------------
static void cleanup(int error)
{
	switch ( error )
	{
		// kompletni uklid pri exitu
		case ModuleExit:
			// zrusime adresar modulu v /sys
			device_destroy(dskel.sys_class_ptr, dskel.dev_num);

		case Failed_SysDev:
			// zrusime tridu
			class_destroy(dskel.sys_class_ptr);

		case Failed_SysClass:
			// vyjmeme modul z jadra
			// pozn.: cdev_del() se vola jen po uspesnem cdev_add()
			cdev_del(dskel.cdev_ptr);

		case Failed_CdevAdd:
			// vypneme casovac (pokud nebezel nic se nestane)
			timer_off();
			// zrusime mapovani portu
			ioport_unmap(dskel.port_base);

		case Failed_IoMap:
			// zrusime rezervaci I/O portu
			release_region(PAR_PORT_BASE, PAR_PORT_SIZE);

		case Failed_ReqRegion:
			// pozn.: neexistuje opak fce cdev_alloc(), uvolneni pameti resi
			// jadro samo
		case Failed_CdevAlloc:
			// uvolnime nami zabrany major-minor par
			unregister_chrdev_region(dskel.dev_num, DEV_COUNT);

		case Failed_ChrdevRegion:
			// jeste nic se nealokovalo, takze neni co vracet
			break;

		default:
			break;
	}
}

// -----------------------------------------------------------------------------
static int __init dskel_init(void)
{
	int result = 0;
	struct resource * registr_ptr;

	// registrace k major-minor paru
	result = alloc_chrdev_region(&dskel.dev_num, MINOR(dskel.dev_num),
								 DEV_COUNT, DEV_NAME);
	if ( result < 0 )
	{
		cleanup(Failed_ChrdevRegion);
		goto End;
	}

	// alokace pameti pro strukturu cdev
	dskel.cdev_ptr = cdev_alloc();
	if ( dskel.cdev_ptr == NULL )
	{
		result = -ENOMEM;
		cleanup(Failed_CdevAlloc);
		goto End;
	}

	// nastaveni vlastnika a seznamu souborovych operaci
	dskel.cdev_ptr->owner = THIS_MODULE;
	dskel.cdev_ptr->ops = &dskel_file_ops;

	// rezervujeme si paralelni port, resp. jeho I/O port
	registr_ptr = request_region(PAR_PORT_BASE, PAR_PORT_SIZE, DEV_NAME);
	if ( registr_ptr == NULL )
	{
		pr_err("dskel_init(): parallel port already in use!\n");
		cleanup(Failed_ReqRegion);
		result = -EBUSY;
		goto End;
	}

	// premapujeme port jako I/O pamet
	dskel.port_base = ioport_map(PAR_PORT_BASE, PAR_PORT_SIZE);
	if ( dskel.port_base == NULL )
	{
		cleanup(Failed_IoMap);
		result = -ENOMEM;
		goto End;
	}

	// inicializace casovace (zbytek se nastavuje pri spusteni casovace)
	dskel.timer.function = &timer_userfunction;
	dskel.timer.data = (unsigned long) NULL;
	init_timer(&dskel.timer);

	//  --- pridani modulu do jadra - uz lze volat fce modulu ---
	result = cdev_add(dskel.cdev_ptr, dskel.dev_num, DEV_COUNT);
	if ( result < 0 )
	{
		cleanup(Failed_CdevAdd);
		goto End;
	}

	// vytvorime si tridu v /sys/class
	dskel.sys_class_ptr = class_create(THIS_MODULE, DEV_NAME);
	if ( IS_ERR(dskel.sys_class_ptr) )
	{
		cleanup(Failed_SysClass);
		result = PTR_ERR(dskel.sys_class_ptr);
		goto End;
	}

	// pridame nas modul do /sys/class/dskel
	dskel.sys_dev_ptr = device_create(dskel.sys_class_ptr,
									  NULL, dskel.dev_num, NULL, DEV_NAME);
	if ( IS_ERR(dskel.sys_dev_ptr) )
	{
		cleanup(Failed_SysDev);
		result = PTR_ERR(dskel.sys_dev_ptr);
		goto End;
	}

End:
	pr_alert("dskel_init(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static void __exit dskel_exit(void)
{
	cleanup(ModuleExit);

	pr_alert("dskel_exit(): done\n");
}

// -----------------------------------------------------------------------------
module_init(dskel_init);
module_exit(dskel_exit);

MODULE_AUTHOR("mBoss <marek.laznicka@elvoris.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Simple but useful device drive skeleton");
MODULE_VERSION("0.0.1");
